unit main;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, ComCtrls, ExtCtrls,
  ShellCtrls, Grids, Parser, textstream;

const
    parts_path = '\SCAD parts';
    MAX_ALIASES = 500;

type


  tpoint = record
    x0: string;
    y0: string;
    x1: string;
    y1: string;
    curve: string;
    used: boolean;
  end;

  t_validpoint = record
    point:     real;
    valid:     boolean;
  end;

  t_realpoint = record
    x0: t_validpoint;
    y0: t_validpoint;
    x1: t_validpoint;
    y1: t_validpoint;
    curve: t_validpoint;
    used: boolean;
  end;

  thole = record
    diameter: real;        // diameter
    x: real;        // relative x as designed
    y: real;        // relative y as designed
  end;



  // enumerated types for alias lookups
  // add here to build enumerated type for actual drawing

  part_type_type = (part_unknown, part_transistor, part_passive,
    part_diode, part_resistor, part_capacitor,
    part_led, part_IC, part_hardware,
    part_pot, part_inductor, part_module);

         // rotate types for individual hole positioning

  rotate_type = (R0, R90, R180, R270, MR0, MR90, MR180, MR270);



  { TForm1 }

  TForm1 = class(TForm)
    board_grid_1: TStringGrid;
    board_outline_image: TImage;
    board_outline_StringGrid: TStringGrid;
    board_image1: TImage;
    PageControl2: TPageControl;
    Panel1: TPanel;
    FILES_BUTTON: TPanel;
    Panel3: TPanel;
    Panel6: TPanel;
    Panel7: TPanel;
    PROCESS_ALL_BUTTON: TPanel;
    ShellListView1: TShellListView;
    ShellTreeView1: TShellTreeView;
    StatusBar1: TStatusBar;
    TabSheet3: TTabSheet;
    TabSheet4: TTabSheet;
    TabSheet5: TTabSheet;
    //procedure FILESl_BUTTONClick(Sender: TObject);
    procedure FILES_BUTTONClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure PROCESS_ALL_BUTTONClick(Sender: TObject);
    procedure ShellListView1Change(Sender: TObject; Item: TListItem;
      Change: TItemChange);
    procedure ShellListView1Click(Sender: TObject);
    procedure ShellListView1DblClick(Sender: TObject);
    procedure ShellListView1SelectItem(Sender: TObject; Item: TListItem;
      Selected: Boolean);
    procedure ShellTreeView1Change(Sender: TObject; Node: TTreeNode);
    procedure get_board_size();
    procedure create_files(sourcename: string);
    procedure board_point_sort();
  private

  public
    // file open and select variables

    select_filename:       string;
    select_filepath:       string;

    boardfile:               ttextstream;
    board_filename:          string;

    // board image
    board_outline:           array[0..40] of t_realpoint;             // strings parsed from board
    sorted_outline:          array[0..40] of t_realpoint;        // reals sorted
    board_maxx:              real;
    board_maxy:              real;
    board_minx:              real;
    board_miny:              real;
    // parser
    token: array [0..127] of string [128];
  end;

var
  Form1:                   TForm1;


implementation

{$R *.lfm}


 // *************************************************************************
 // ************************** form *****************************************
 // *************************************************************************


{ TForm1 }

procedure TForm1.FormCreate(Sender: TObject);

begin
   // inits and creates
   boardfile := ttextstream.Create;
   select_filename := '';

   // initial buttons

   process_all_button.Enabled:= false;;
   //select_pathname := '';

   // board page
   board_grid_1.Cells[0,1] := 'Max X';
   board_grid_1.Cells[0,2] := 'Max Y';
   board_grid_1.Cells[0,3] := 'Min X';
   board_grid_1.Cells[0,4] := 'Min Y';

   form1.TabSheet3.TabVisible:= true;
   form1.TabSheet4.TabVisible:= true;
   form1.TabSheet5.TabVisible:= true;
   form1.TabSheet3.Visible:= true;
   form1.TabSheet4.Visible:= true;
   form1.TabSheet5.Visible:= true;
   form1.TabSheet3.Enabled:= true;
   form1.TabSheet4.Enabled:= true;
   form1.TabSheet5.Enabled:= true;

end;

procedure TForm1.FILES_BUTTONClick(Sender: TObject);
begin
  if (select_filename <> '') then
  begin
    process_all_button.Enabled := true;
    boardfile.Clear;
    boardfile.LoadFromFile(board_filename);
  end
  else
  begin
    process_all_button.Enabled := false;
    form1.TabSheet3.TabVisible:= true;
    form1.TabSheet4.TabVisible:= true;
    form1.TabSheet5.TabVisible:= true;
    form1.TabSheet3.Visible:= true;
   form1.TabSheet4.Visible:= true;
   form1.TabSheet5.Visible:= true;

    form1.PageControl2.ActivePage := tabsheet5;
  end;
end;

procedure TForm1.PROCESS_ALL_BUTTONClick(Sender: TObject);
begin
  get_board_size;

end;

procedure TForm1.ShellListView1Change(Sender: TObject; Item: TListItem;
  Change: TItemChange);
begin

end;

//procedure TForm1.FILES_BUTTONClick(Sender: TObject);
//begin
//  pagecontrol1.ActivePage := TabSheet2;
//     if (select_filename <> '') then
//     begin
//       process_all_button.Enabled := true;
//       boardfile.Clear;
//       boardfile.LoadFromFile(board_filename);
//     end
//     else
//     begin
//       process_all_button.Enabled := false;
//     end;
//end;

procedure TForm1.ShellListView1Click(Sender: TObject);
begin

end;


// DOUBLE CLICK ON FILENAME, OPEN IT IF *.BRD
procedure TForm1.ShellListView1DblClick(Sender: TObject);
var
  filename:             string;
  extension:            string;
begin
  extension := extractfileExt(select_filename);
  if (extension = '.brd') then
  begin
    filename := select_filepath + '/' +select_filename;
    statusbar1.Panels[0].Text:= filename;
    board_filename := filename;
  end;
end;

procedure TForm1.ShellListView1SelectItem(Sender: TObject; Item: TListItem;
  Selected: Boolean);
begin
   select_filename := item.Caption;
end;

procedure TForm1.ShellTreeView1Change(Sender: TObject; Node: TTreeNode);
begin
  select_filepath := Node.GetTextPath;
end;



// **********************************************************************************************************
// ***************************************** board point sorting ********************************************
// **********************************************************************************************************



// board has board pairs of points, x0,y0 and x1,y1 and curve
// order points so that x0,y0 of next point is equal to x1,y1 of previous point


procedure Tform1.board_point_sort();
var
  i, j: integer;
  done: boolean;
  min_x, max_x:                          real;
  min_y, max_y:                          real;
  current_segment:                       integer;
  search_segment:                        integer;
  matched_segment:                       integer;
  sorted_list:                           array[0..40] of integer;
  available_list:                        array[0..40] of integer;


  // ****************************** find next point ************************************************************

  // always start at 1
  // look for first valid
  function find_next(searching: integer): integer;
  begin
    inc(searching);
    while (searching <= 40) and (board_outline[searching].used = true) do
    begin
      inc(searching);
    end;
    if searching > 40 then
    begin
      result:= -1;
    end
    else
    begin
      result := searching;
    end;
  end;

// ************************************ match end segment to beginning next *******************************************

  // match beginning to end, works for normal order
  // uses supplied segment cseg,

  function match_end_to_beginning(cseg: integer): integer;
  var
    sseg:                     integer;
    done:                     boolean;

  begin

    sseg := 0;
    //  at first, will find first unused segment
    sseg := find_next(sseg);            // first unused segment in board outline
    if sseg = -1 then                   // out of segments, all used
    begin
      result := -1;                     // out of rope
    end
    else
    begin
      done := false;
      while (not done) and (sseg <> -1) do
      begin
        //board outline end compare to real_outline beginning
        if (sorted_outline[cseg].x1.point = board_outline[sseg].x0.point) and
           (sorted_outline[cseg].y1.point = board_outline[sseg].y0.point) then
        begin
          // found match
          done := true;
          result := sseg;
          board_outline[sseg].used:= true;
        end
        else
        begin
          // no match, look at next up
          sseg := find_next(sseg);
        end;
      end;
    end;
    result := sseg;
  end;

// **************************** match end segment to end segment ********************************************

  //// match end to end, works for reversed segment
  //function match_end_to_end: integer;
  //var
  //  cseg,sseg:                     integer;
  //  done:                          boolean;
  //
  //begin
    //sseg := 1;
    //done := false;
    //cseg := current_segment;
    //
    //while (real_outline[cseg].used = true) and (cseg < 40) and (done = false) do
    //begin
    //  done := false;
    //  sseg := cseg + 1;
    //  while (not done) and (sseg <= 40) do
    //  begin
    //    // real outline end compare to real_outline beginning
    //    if (real_outline[cseg].x1 = real_outline[sseg].x1) and
    //       (real_outline[cseg].y1 = real_outline[sseg].y1) then
    //    begin
    //      // found match
    //      done := true;
    //    end
    //    else
    //    begin
    //      // no match, look at next up
    //      inc(sseg);
    //    end;
    //  end;
    //  if (done = true) then
    //  begin
    //    // successful match
    //    // store and update
    //    result := sseg;
    //  end
    //  else
    //  begin
    //    // no match
    //    result := -1;
    //  end;
    //end;
  //end;

  //board_outline:   tboardoutline;         // strings
  //sorted_outline:  tboardoutline;         // strings
  //real_outline:    tboardreal;            // real numbers
  //real_outline:     tboardreal;            // REAL Numbers

  // **************************** beginning board outline *************************************************

begin
  // sorted outline will be returned as real numbers
  for i := 0 to 40 do
  begin
    sorted_outline[i].used := False;
    sorted_outline[i].x0.point:= 0.0;
    sorted_outline[i].y0.point:= 0.0;
    sorted_outline[i].x1.point:= 0.0;
    sorted_outline[i].y1.point:= 0.0;
    sorted_outline[i].curve.point:= 0.0;
    board_outline[i].used:= false;
  end;


  // real_outline contains line segments, not sorted
  // sorted_outline has (gradually) sorted segments

  // sorted list's first segment has to be available list first segment
  // start with first segment to compare against
  current_segment := 0;
  board_outline[current_segment].used:= true;             // segment #1 is already in list
  sorted_outline[current_segment] := board_outline[current_segment];

  // and do for all available segments, current segment points to sorted_list
  // at beginning of list, compare [1-n] against [0]
  while (current_segment < 40) and (board_outline[current_segment].x0.valid = true) do
  begin
    // look at next available segment in list (match does all) and match current segment
    // look at current end to match to beginning segment
    matched_segment := match_end_to_beginning(current_segment);
    if (matched_segment > 0) then
    begin
      inc(current_segment);                                               // next slot
      sorted_outline[current_segment] := board_outline[matched_segment];  // copy data
      board_outline[matched_segment].used:= true;                         // flag used
    end
    else
    begin
      inc(current_segment);
    end;
  end;
end;



// **********************************************************************************************************
// ***************************************** get board size *************************************************
// **********************************************************************************************************

// scans for board size, and builds board outline data
// does not create output, sets up board_outline by points
// creates point array for hull
// MODIFIED 1/19/21

procedure Tform1.get_board_size();
var
  done:         boolean;
  buffer:       string;
  board_found:  boolean;
  side_count:   integer;
  x1string:     string;
  y1string:     string;
  x2string:     string;
  y2string:     string;
  xtemp:        real;
  ytemp:        real;
  curve:        string;
  display_height:       integer;
  display_width:        integer;

  xmax, ymax:   real;
  xmin, ymin:   real;
  x,y:          real;
  j,i:          integer;
  endmarker:    integer;

  // added
  layer:        real;

  // look for
  // <wire x1="6.35" y1="0" x2="0" y2="5.715" width="0.254" layer="20" curve="-96.340192"/>
  // <wire x1="90.93" y1="98.52" x2="0" y2="98.52" width="0.254" layer="20"/>

  function find_wire(): boolean;
  var
    iii:               integer;
    found:             boolean;

  begin
    iii := 0;
    found := false;
    find_wire := false;
    while ((not found) and (iii < 127)) do
    begin
      if (token[iii] = 'wire') then
      begin
        found := true;
        find_wire := true;
      end
      else
      begin
        inc (iii);
      end;
    end;
  end;

  function decode_token(name: string; var value: real): boolean;
  var
    ii:                       integer;
    done:                     boolean;

  begin
    done := false;
    ii := 0;
    decode_token := false;
    while (not done) and (token[ii] <> '') and (ii < 128) do
    begin
      if (upcase(token[ii]) = upcase(name)) then
      begin
        if (token[ii+1] = '=') then
        begin
          // token[ii+2] is either +, -, or #
          case (token[ii+2]) of
            '-': begin
              value := -strtofloat(token[ii+3]);
              done := true;
              decode_token := true;
            end;
            '+': begin
              value := strtofloat(token[ii+3]);
              done := true;
              decode_token := true;
            end;
            else
            begin
              value := strtofloat(token[ii+2]);
              decode_token := true;
              done := true;
            end;
          end;
        end
      end
      else
      begin
        inc(ii);
      end;
    end;
  end;



  // NOTE:
  //    graphics coordinates are at 0,0 top left
  //    board coordinates are at 0,height (bottom left)

begin
  // initialize
  for i := 0 to 40 do
  begin
    board_outline[i].used := False;
    board_outline[i].x0.point:= 0.0;
    board_outline[i].y0.point:= 0.0;
    board_outline[i].x1.point:= 0.0;
    board_outline[i].y1.point:= 0.0;
    board_outline[i].curve.point:= 0.0;

  end;

  display_height := board_image1.Height;
  display_width := board_image1.Width;
  // board is assumed to be mostly positive coordinates, but must be greater than -1000
  xmax := -1000.0;
  ymax := -1000.0;
  xmin := 1000.0;
  ymin := 1000.0;
  done := False;
  side_count := 0;
  board_found := False;

  // parse board file
  boardfile.seek(0, sofrombeginning);
  boardfile.EOF:= false;
  board_outline_stringgrid.Clean(0,1,6,40,[gzNormal]);
  while ((not boardfile.EOF) and (not done)) do
  begin

    // read line and parse it
    buffer := '';
    boardfile.readln(buffer);
    if (buffer = '') then done := true;
    if (not done) then
    begin
      parse(buffer,token);

      // check for <board> in file, no further entries on line
      // <board> sets board_found, </board> at very end of file
      if ((token[0] = '<') and (token[1] = 'board')) then
      begin
        board_found := True;
      end;
    //  // begin board section
    //
    //  // board data for outline starts with <plain>
    //  // and ends with </plain>
    //  // loop end detect
      if (board_found = true) then
      begin
        if ((token[0] = '<') and (token[1] = '/plain')) then
        begin
          done := True;
        end;
        if ((token[0] = '<') and (token[1] = '/board')) then
        begin
          done := True;
        end;
        // look for
        // <wire x1="6.35" y1="0" x2="0" y2="5.715" width="0.254" layer="20" curve="-96.340192"/>
        // <wire x1="90.93" y1="98.52" x2="0" y2="98.52" width="0.254" layer="20"/>
        if (find_wire) then
        begin
          // look for layer, if layer = 20 then look for outline
          if (decode_token('layer', layer)) then
          begin
            if (layer = 20) then
            begin
              // layer = 20, therefore dimension layer
              // look for x1
              if (decode_token('x1', board_outline[side_count].x0.point)) then
              begin
                board_outline[side_count].x0.valid:= true;
              end;
              // look for y1
              if (decode_token('y1', board_outline[side_count].y0.point)) then
              begin
                board_outline[side_count].y0.valid:= true;
              end;
              // look for x2
              if (decode_token('x2', board_outline[side_count].x1.point)) then
              begin
                board_outline[side_count].x1.valid:= true;
              end;
              // look for y2
              if (decode_token('y2', board_outline[side_count].y1.point)) then
              begin
                board_outline[side_count].y1.valid:= true;
              end;
              // look for curve
              if (decode_token('curve', board_outline[side_count].curve.point)) then
              begin
                board_outline[side_count].curve.valid:= true;
              end;
              if (board_outline[side_count].x0.valid) and
                 (board_outline[side_count].y0.valid) and
                 (board_outline[side_count].x1.valid) and
                 (board_outline[side_count].x0.valid) then
              begin
                board_outline[side_count].used := True;
                board_outline[side_count+1].used := False;
                if (board_outline[side_count].x0.point > xmax) then xmax := board_outline[side_count].x0.point;
                if (board_outline[side_count].y0.point > ymax) then ymax := board_outline[side_count].y0.point;

                if (board_outline[side_count].x0.point < xmin) then xmin := board_outline[side_count].x0.point;
                if (board_outline[side_count].y0.point < ymin) then ymin := board_outline[side_count].y0.point;
                board_outline_stringgrid.Cells[0,1+side_count] := inttostr(side_count);
                board_outline_stringgrid.Cells[1,1+side_count] := floattostr(board_outline[side_count].x0.point);
                board_outline_stringgrid.Cells[2,1+side_count] := floattostr(board_outline[side_count].y0.point);
                board_outline_stringgrid.Cells[3,1+side_count] := floattostr(board_outline[side_count].x1.point);
                board_outline_stringgrid.Cells[4,1+side_count] := floattostr(board_outline[side_count].y1.point);
                board_outline_stringgrid.Cells[5,1+side_count] := floattostr(board_outline[side_count].curve.point);
                inc (side_count);
              end;            // end valid point
            end;              // end layer = 20
          end;                // end layer
        end;                  // end wire
      end;                    // end in board
    end;                      // not done (read null buffer)
  end;                      // end scan of file

  // process
  board_point_sort;

  // record this

  board_maxx := xmax;
  board_maxy := ymax;
  board_minx := xmin;
  board_miny := ymin;
  board_grid_1.Cells[1,1] := floattostr(xmax);
  board_grid_1.Cells[1,2] := floattostr(ymax);
  board_grid_1.Cells[1,3] := floattostr(xmin);
  board_grid_1.Cells[1,4] := floattostr(ymin);

  // display this
  board_image1.Canvas.Brush.Color:= clblack;
  board_image1.Canvas.FillRect(board_image1.Canvas.ClipRect);

  board_image1.Canvas.Pen.Color:= clwhite;
  for i := 0 to 40 do
  begin
    if (board_outline[i].x0.valid = true) then
    begin
      if (board_outline[i].curve.valid = true) then
      begin
        //board_image1.Canvas.ARC(

        //trunc(30 + board_outline[i].x0.point * 10),
        //display_height - trunc(30 + board_outline[i].y0.point * 10),
        //trunc(30 + board_outline[i].x1.point * 10),
        //display_height - trunc(30 + board_outline[i].y1.point * 10));
        //

        //if (board_outline[i].x0.point > board_outline[i].x1.point) then
        //begin
        //  // left half
        //  if(board_outline[i].y0.point > board_outline[i].y1.point) then
        //  begin
        //    // lower left
        //    board_image1.Canvas.AngleArc(
        //    trunc(30 + board_outline[i].x0.point * 10),
        //    display_height - trunc(30 + board_outline[i].y1.point * 10),
        //    trunc(30 + board_outline[i].x0.point * 10 - board_outline[i].x1.point * 10),
        //    90,90);
        //
        //  end
        //  else
        //  begin
        //    // upper left
        //  end;
        //end
        //else
        //begin
        //  // right half
        //  if(board_outline[i].y0.point > board_outline[i].y1.point) then
        //  begin
        //    // lower right
        //  end
        //  else
        //  begin
        //    // upper right
        //  end;
        //end;


        board_image1.Canvas.Line(
        trunc(30 + board_outline[i].x0.point * 10),
        display_height - trunc(30 + board_outline[i].y0.point * 10),
        trunc(30 + board_outline[i].x1.point * 10),
        display_height - trunc(30 + board_outline[i].y1.point * 10));
      end
      else
      begin
        board_image1.Canvas.Line(
        trunc(30 + board_outline[i].x0.point * 10),
        display_height - trunc(30 + board_outline[i].y0.point * 10),
        trunc(30 + board_outline[i].x1.point * 10),
        display_height - trunc(30 + board_outline[i].y1.point * 10));
      end;
    end;
  end;

  //geometry.board_maxx := xmax;
  //geometry.board_maxy := ymax;

end;




 
// ********************************************************************************************************
// ********************************************************************************************************
// **************************************** file creates, writes, outputs *********************************
// ********************************************************************************************************
// ********************************************************************************************************

procedure Tform1.create_files(sourcename: string);
var
  filepath: string;
  filebase: string;
  i: integer;
begin
  // display stuff
  //parts_used_list.Clear;
  //not_found_list1.Clear;
  //Found_list.Clear;
  // file and stream clears
  //outputfile.Clear;
  boardfile.Clear;
  //punchlist.Clear;
  //postlist.Clear;
  //boardsizefile.Clear;
  //holefile.Clear;
  //punchfile.Clear;
  //postfile.Clear;
  // filename creates and file read from board file
  filebase := ExtractFileName(sourcename);
  for i := 1 to length(filebase) - 1 do
  begin
    if filebase[i] = '-' then
      filebase[i] := '_';
    if filebase[i] = '.' then
      filebase[i] := '_';
    if filebase[i] = ' ' then
      filebase[i] := '_';
  end;
  //projectname := filebase;
  //projectname := changefileext(projectname, '');
  //boardfile.readtextfromfile(sourcename);
  //StatusBar1.Panels[0].Text := sourcename;
  //filepath := ExtractFilePath(sourcename);
  //outfilename := changefileext(projectname, '.scad');
  //
  //outfilename := filepath + outfilename;
  //punchfilename := outfilename;
  //punchfilename := changefileext(punchfilename, '');
  //punchfilename := punchfilename + '_punch.scad';

  //postfilename := outfilename;
  //postfilename := changefileext(postfilename, '');
  //postfilename := postfilename + '_post.scad';

  // these write to files
  //outputposttext(sourcename);                     // project header and includes
  //outputpostBeginningText();                     // definition of parts module

  // now make files, these write to stream
  //
  get_board_size();
  // does not create a file, extracts board data
  //build_parts();
  // creates stream of parts and hole definitions
  //build_board_and_holes();
  // these have to be after build parts because board text generates uses file

  // these write to files
  //outputboardtext(sourcename);                    // project header and includes
  //outputpunchtext(sourcename);

  //outputPartsBeginningText();                     // definition of parts module
  //outputpunchBeginningText();

  // write any updates from stream
  //partsfile.SaveToStream(outputfile);             // write parts created from file
  //punchlist.SaveToStream(punchfile);
  //postlist.SaveToStream(postfile);

  //OutputPartsEndingText();                        // end text of parts module
  //boardsizefile.SaveToStream(outputfile);

  //OutputPunchEndingText();                        // end text of parts module
  //OutputPostEndingText();                        // end text of parts module

  // write board and holes created from file

  //outputfile.writeln(sanitize(projectname) + '_board();');
  // at end of file draws board
  //outputfile.writeln(sanitize(projectname) + '_parts();');
  // at end of file draws parts
  //punchfile.writeln(sanitize(projectname) + '_punch();');
  // at end of file draws parts
  //postfile.writeln(sanitize(projectname) + '_post();');
  // at end of file draws parts


  //outputfile.writetexttofile(outfilename);
  //punchfile.writetexttofile(punchfilename);
  //postfile.writetexttofile(postfilename);

end;

















end.



