unit textstream;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils;


type
ttextstream = class(tmemorystream)
  EOF: boolean;
  where: longint;
  constructor Create;
  procedure erase;
  procedure writetexttofile(filename: string);
  procedure readtextfromfile(filename: string);
  function readln(var inbuf: string): integer;
  procedure writeln(outbuf: string);
end;


implementation

// **********************************************************************************************************
// ***************************************** TTEXTSTREAM routines *******************************************
// **********************************************************************************************************


{ throws away terminating characters such as lf or cr. }

constructor ttextstream.Create;
begin
  inherited Create;
  EOF := False;
  where := 0;
end;

procedure ttextstream.erase;
begin
  self.Clear;
  EOF := False;
end;


// return number of characters read

function ttextstream.readln(var inbuf: string): integer;

var
  i: integer;
  inchar: byte;
  doneline: boolean;

begin
  doneline := False;
  i := 0;
  readln := 0;
  inbuf := '';
  inchar := 0;
  { reads cr, does throws away, reads lf }
  while (i < 255) and (inchar <> $0a) and (not doneline) do
  begin
    if Read(inchar, 1) <> 1 then
    begin
      doneline := True;
      readln := i;
    end
    else
    begin
      if (inchar = $0d) then
      begin
        doneline := True;
      end
      else
      begin
        if (self.Position > self.Size) then
        begin
          doneline := True;
        end
        else
        begin
          inbuf := inbuf + chr(inchar);
          Inc(i);
          readln := i;
        end;
      end;
    end;
  end;

  where := position;
  Read(inchar, 1);
  case inchar of
    $0:
    begin
      readln := -1;
    end;
    $0D:
    begin
      where := position;
    end;
    $0A:
    begin
      where := position;
    end;
    else    { read character, but is not cr or lf, so reset position }
      seek(-1, sofromcurrent);
  end;
  if (position = size) then
    EOF := True
  else
    EOF := False;
end;

procedure ttextstream.writeln(outbuf: string);

var
  cr: char;
  lf: char;

begin
  cr := #$0D;
  lf := #$0A;
  {  inc(currentline[currentinclude]);}
  //  index[currentline] := self.Size;
  outbuf := trimright(outbuf);
  { default output string length was set to 255, and padded }
  if length(outbuf) <> 0 then
  begin
    Write(outbuf[1], length(outbuf));
    Write(cr, 1);
    Write(lf, 1);
  end
  else
  begin
    Write(cr, 1);
    Write(lf, 1);
  end;
end;

procedure ttextstream.writetexttofile(filename: string);
begin
  seek(0, sofrombeginning);
  self.SaveToFile(filename);
end;

procedure ttextstream.readtextfromfile(filename: string);
begin
  self.LoadFromFile(filename);
  seek(0, sofrombeginning);
end;



end.

